//this file is part of eMule
//Copyright (C)2002 Merkur ( merkur-@users.sourceforge.net / http://www.emule-project.net )
//
//This program is free software; you can redistribute it and/or
//modify it under the terms of the GNU General Public License
//as published by the Free Software Foundation; either
//version 2 of the License, or (at your option) any later version.
//
//This program is distributed in the hope that it will be useful,
//but WITHOUT ANY WARRANTY; without even the implied warranty of
//MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//GNU General Public License for more details.
//
//You should have received a copy of the GNU General Public License
//along with this program; if not, write to the Free Software
//Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#include "stdafx.h"
#include "DebugHelpers.h"
#include "emule.h"
#include "ListenSocket.h"
#include "opcodes.h"
#include "UpDownClient.h"
#include "ClientList.h"
#include "OtherFunctions.h"
#include "DownloadQueue.h"
#include "IPFilter.h"
#include "SharedFileList.h"
#include "PartFile.h"
#include "SafeFile.h"
#include "Packets.h"
#include "UploadQueue.h"
#include "ServerList.h"
#include "Server.h"
#include "Sockets.h"
#include "BandWidthControl.h" // [TPT]
#ifndef _CONSOLE
#include "emuledlg.h"
#include "TransferWnd.h"
#include "ClientListCtrl.h"
#include "ChatWnd.h"
#endif

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif


// CClientReqSocket
CClientReqSocket::CClientReqSocket(CPreferences* in_prefs,CUpDownClient* in_client){
	app_prefs = in_prefs;
	client = in_client;
	if (in_client)
		client->socket = this;
	theApp.listensocket->AddSocket(this);
	ResetTimeOutTimer();
	deletethis = false;
	deltimer = 0;
}

// [TPT]
CClientReqSocket::~CClientReqSocket(){
	// Maella -Code Fix-
	if(client != NULL && client->socket == this){
		client->socket = NULL;
	}
	client = NULL;
	// Maella end
	theApp.listensocket->RemoveSocket(this);

	DEBUG_ONLY (theApp.clientlist->Debug_SocketDeleted(this));
}

void CClientReqSocket::ResetTimeOutTimer(){
	timeout_timer = ::GetTickCount();
}

// [TPT] - Maella  -Unnecessary Protocol Overload-
bool CClientReqSocket::CheckTimeOut()
{
	UINT uTimeout = CONNECTION_TIMEOUT;
	if(client && client->GetChatState()!=MS_NONE &&
	  (client->GetUploadState() == US_ONUPLOADQUEUE || client->GetUploadState() == US_NONE) &&
	  (client->GetDownloadState() == DS_ONQUEUE     || client->GetDownloadState() == DS_NONE))
		uTimeout += CONNECTION_TIMEOUT;
	if (::GetTickCount() - timeout_timer > uTimeout){
		timeout_timer = ::GetTickCount();
		Disconnect();
		return true;
	}
	return false;
}
// Maella end

void CClientReqSocket::OnClose(int nErrorCode){
	ASSERT (theApp.listensocket->IsValidSocket(this));
	CEMSocket::OnClose(nErrorCode);
	Disconnect();
}

void CClientReqSocket::Disconnect(){
	AsyncSelect(0);
	byConnected = ES_DISCONNECTED;
	if (!client)
		Safe_Delete();
	else
		if(client->Disconnected(true)){
			CUpDownClient* temp = client;
			client->socket = NULL;
			client = NULL;
			delete temp;
			Safe_Delete();
		}
		else{
			client = NULL;
			Safe_Delete();
		}
};

void CClientReqSocket::Delete_Timed(){
// it seems that MFC Sockets call socketfunctions after they are deleted, even if the socket is closed
// and select(0) is set. So we need to wait some time to make sure this doesn't happens
	if (::GetTickCount() - deltimer > 10000)
		delete this;
}

// [TPT] - Maella -Code Fix-
void CClientReqSocket::Safe_Delete(){
	// Two persons have reported here some strange errors
	// I'm not sure how it happened
	ASSERT (theApp.listensocket->IsValidSocket(this));
	AsyncSelect(0);
	deltimer = ::GetTickCount();
	if (m_SocketData.hSocket != INVALID_SOCKET) // deadlake PROXYSUPPORT - changed to AsyncSocketEx
		ShutDown(SD_BOTH);
	byConnected = ES_DISCONNECTED;
	deletethis = true;

	if(client != NULL && client->socket == this){
		// They might have an error in the cross link somewhere
		client->socket = NULL;
	}
	client = NULL;
}
// Maella end

bool CClientReqSocket::ProcessPacket(char* packet, uint32 size, UINT opcode){
	try{
		//try{ // [TPT]
			if (!client && opcode != OP_HELLO){
				theApp.downloadqueue->AddDownDataOverheadOther(size);
				throw GetResString(IDS_ERR_NOHELLO);
			}
			switch(opcode){
				//eWombat [SNAFU][TPT]
				case OP_RENNSCHWEIN:{
					if (client)
						client->DoSnafu(SNAFU_OPCODE,false);
					break;
}
				//SNAFU END
				case OP_HELLOANSWER:{
					theApp.downloadqueue->AddDownDataOverheadOther(size);
					client->ProcessHelloAnswer(packet,size);

					// start secure identification, if
					//  - we have received OP_EMULEINFO and OP_HELLOANSWER (old eMule)
					//	- we have received eMule-OP_HELLOANSWER (new eMule)
					if (client->GetInfoPacketsReceived() == IP_BOTH)
						client->InfoPacketsReceived();

					if (client){
						client->ConnectionEstablished();
						theApp.emuledlg->transferwnd->clientlistctrl.RefreshClient(client);
					}
					break;
				}
				case OP_HELLO:{
					theApp.downloadqueue->AddDownDataOverheadOther(size);

				// Maella -Code Fix-
				if(client != NULL){
					CString oldHash;
					CString newHash;
					if(size > 17){
						oldHash = (client->GetUserHash() == NULL) ? _T("null") : EncodeBase16(client->GetUserHash(), 16);
						newHash = EncodeBase16((const uchar*)&packet[1], 16);
					}
					if(oldHash != newHash){
						AddLogLine(true,  _T("User %s (client=%s) try to send multiple OP_HELLO, old hash=%s, new has=%s"), client->GetUserName, client->GetClientSoftVer(), oldHash, newHash);
					}
					else {
						AddLogLine(true, _T("User %s (client=%s) try to send multiple OP_HELLO"), client->GetUserName, client->GetClientSoftVer());
					}
					throw CString(_T("Invalid request received"));
				}
				// Maella end

					bool bNewClient = !client;
					if (bNewClient){
						// create new client to save standart informations
						client = new CUpDownClient(this);
					}

					bool bIsMuleHello = false;
					try{
						bIsMuleHello = client->ProcessHelloPacket(packet,size);
					}
					catch(...){
						if (bNewClient){
							// Don't let CUpDownClient::Disconnected be processed for a client which is not in the list of clients.
							delete client;
							client = NULL;
						}
						throw;
					}

					// if IP is filtered, dont reply but disconnect...
					if (theApp.ipfilter->IsFiltered(client->GetIP())) {
						if (theApp.glob_prefs->GetLogFilteredIPs() == false)
							AddDebugLogLine(true,GetResString(IDS_IPFILTERED),client->GetFullIP(),theApp.ipfilter->GetLastHit());
						theApp.stat_filteredclients++;
						if (bNewClient){
							delete client;
							client = NULL;
						}
						Disconnect();
						return false;
					}

					// now we check if we know this client already. if yes this socket will
					// be attached to the known client, the new client will be deleted
					// and the var. "client" will point to the known client.
					// if not we keep our new-constructed client ;)
					if (theApp.clientlist->AttachToAlreadyKnown(&client,this)){
						// update the old client informations
						bIsMuleHello = client->ProcessHelloPacket(packet,size);						
					}
					else {
						theApp.clientlist->AddClient(client);
						client->SetCommentDirty();
					}

					theApp.emuledlg->transferwnd->clientlistctrl.RefreshClient(client);

					// send a response packet with standart informations
					if (client->GetHashType() == SO_EMULE && !bIsMuleHello)
						client->SendMuleInfoPacket(false);

					client->SendHelloAnswer();
					if (client)
						client->ConnectionEstablished();

					// start secure identification, if
					//	- we have received eMule-OP_HELLO (new eMule)
					if (client->GetInfoPacketsReceived() == IP_BOTH)
						client->InfoPacketsReceived();
					break;
				}
				case OP_FILEREQUEST:{
					theApp.downloadqueue->AddDownDataOverheadFileRequest(size);
					client->CheckHandshakeFinished(OP_EDONKEYPROT, opcode);

					// IP banned, no answer for this request
					if (client->IsBanned() ){
						break;						
					}

					if (size == 16 || (size > 16 && client->GetExtendedRequestsVersion() > 0)){
						if (!client->GetWaitStartTime())
							client->SetWaitStartTime();
						uchar reqfileid[16];
						md4cpy(reqfileid,packet);
						CKnownFile* reqfile = theApp.sharedfiles->GetFileByID(reqfileid);
						if (!reqfile){
							// if we've just started a download we may want to use that client as a source
							CKnownFile* partfile = theApp.downloadqueue->GetFileByID(reqfileid);
							if (partfile && partfile->IsPartFile())
								if( theApp.glob_prefs->GetMaxSourcePerFile() > ((CPartFile*)partfile)->GetSourceCount() ) //<<--
									theApp.downloadqueue->CheckAndAddKnownSource((CPartFile*)partfile,client);
						}
						if (!reqfile){
							// 26-Jul-2003: removed sending of OP_FILEREQANSNOFIL when receiving OP_FILEREQUEST
							//	for better compatibility with ed2k protocol (eDonkeyHybrid) and to save some traffic
							//
							//	*) The OP_FILEREQANSNOFIL _has_ to be sent on receiving OP_SETREQFILEID.
							//	*) The OP_FILEREQANSNOFIL _may_ be sent on receiving OP_FILEREQUEST but in almost all cases 
							//	   it's not needed because the remote client will also send a OP_SETREQFILEID.

							// DbT:FileRequest
							// send file request no such file packet (0x48)
//							Packet* replypacket = new Packet(OP_FILEREQANSNOFIL, 16);
//							md4cpy(replypacket->pBuffer, packet);
//							theApp.uploadqueue->AddUpDataOverheadFileRequest(replypacket->size);
//							SendPacket(replypacket, true);
							// DbT:End
							break;
						}
						// if wer are downloading this file, this could be a new source
						if (reqfile->IsPartFile())
							if( theApp.glob_prefs->GetMaxSourcePerFile() > ((CPartFile*)reqfile)->GetSourceCount() ) //<<--
								theApp.downloadqueue->CheckAndAddKnownSource((CPartFile*)reqfile,client);
						
						// check to see if this is a new file they are asking for
						if(md4cmp(client->GetUploadFileID(), packet) != 0)
							client->SetCommentDirty();

//						CKnownFile* clientreqfile = theApp.sharedfiles->GetFileByID(client->GetUploadFileID());
//						if( clientreqfile )
//							clientreqfile->SubQueuedCount();
//						reqfile->AddQueuedCount();
						// send filename etc
						client->SetUploadFileID((uchar*)packet);
//						md4cpy(client->reqfileid,packet);
						CSafeMemFile data(128);
						data.Write(reqfile->GetFileHash(),16);
						uint16 namelength = (uint16)strlen(reqfile->GetFileName());
						data.Write(&namelength,2);
						data.Write(reqfile->GetFileName(),namelength);
						// TODO: Don't let 'ProcessUpFileStatus' re-process the entire packet and search the fileid
						// again in 'sharedfiles' -> waste of time.
						client->ProcessUpFileStatus(packet,size);
						Packet* packet = new Packet(&data);
						packet->opcode = OP_FILEREQANSWER;
						theApp.uploadqueue->AddUpDataOverheadFileRequest(packet->size);
						SendPacket(packet,true);
						client->SendCommentInfo(reqfile);
						break;
					}
					throw GetResString(IDS_ERR_WRONGPACKAGESIZE);
					break;
				}
				case OP_FILEREQANSNOFIL:{
					theApp.downloadqueue->AddDownDataOverheadFileRequest(size);
					client->CheckHandshakeFinished(OP_EDONKEYPROT, opcode);

					// DbT:FileRequest
					if (size == 16) {
						// if that client do not have my file maybe has another different
						CPartFile* reqfile = theApp.downloadqueue->GetFileByID((uchar*)packet);
						if (!reqfile)
							break;

						// we try to swap to another file ignoring no needed parts files
						if (client){
							switch (client->GetDownloadState()) {
								case DS_CONNECTED:
								case DS_ONQUEUE:
								case DS_NONEEDEDPARTS:
								if (!client->SwapToAnotherFile(true, true, true, NULL)) {
									theApp.downloadqueue->RemoveSource(client, true);
								}
								break;
							}
						}
						break;
					}
					throw GetResString(IDS_ERR_WRONGPACKAGESIZE);
					break;
					// DbT:End
				}
				case OP_FILEREQANSWER:{
					theApp.downloadqueue->AddDownDataOverheadFileRequest(size);
					client->CheckHandshakeFinished(OP_EDONKEYPROT, opcode);
					client->ProcessFileInfo(packet,size);
					break;
				}
				case OP_FILESTATUS:{
					theApp.downloadqueue->AddDownDataOverheadFileRequest(size);
					client->CheckHandshakeFinished(OP_EDONKEYPROT, opcode);
					client->ProcessFileStatus(packet,size);
					break;
				}
				case OP_STARTUPLOADREQ:{
					theApp.downloadqueue->AddDownDataOverheadFileRequest(size);
					if (!client->CheckHandshakeFinished(OP_EDONKEYPROT, opcode))
						break;
					if( size == 16 ){
						uchar reqfileid[16];
						md4cpy(reqfileid,packet);
						CKnownFile* reqfile = theApp.sharedfiles->GetFileByID(reqfileid);
						if (reqfile){
							if (reqfile->IsPartFile())
								if( theApp.glob_prefs->GetMaxSourcePerFile() > ((CPartFile*)reqfile)->GetSourceCount() ) //<<--
									theApp.downloadqueue->CheckAndAddKnownSource((CPartFile*)reqfile,client);
							if(md4cmp(client->GetUploadFileID(), packet) != 0)
								client->SetCommentDirty();
							client->SetUploadFileID((uchar*)packet);
							client->SendCommentInfo(reqfile);
						}
					}
					theApp.uploadqueue->AddClientToQueue(client);
					break;
				}
				case OP_QUEUERANK:{
					theApp.downloadqueue->AddDownDataOverheadFileRequest(size);
					client->CheckHandshakeFinished(OP_EDONKEYPROT, opcode);
					CSafeMemFile data((BYTE*)packet,size);
					uint32 rank;
					data.Read(&rank,4);
					client->SetRemoteQueueRank(rank);
					break;
				}
				case OP_ACCEPTUPLOADREQ:{
					theApp.downloadqueue->AddDownDataOverheadFileRequest(size);
					client->CheckHandshakeFinished(OP_EDONKEYPROT, opcode);
					if (client->reqfile && !client->reqfile->IsStopped() && 
						(client->reqfile->GetStatus()==PS_READY || client->reqfile->GetStatus()==PS_EMPTY)){
						client->SetSentCancelTransfer(0);
						if (client->GetDownloadState() == DS_ONQUEUE){
							client->SetDownloadState(DS_DOWNLOADING);
							// -khaos--+++> Set our bool to false.  When (if) we receive a block from this client, it will be set to true.
							//				When the download state is changed from DS_DOWNLOADING to something else, we will check this
							//				bool and if it's true and new state != DS_ERROR, then it was a successful session.  Else failed.
							client->InitTransferredDownMini();
							client->SetDownStartTime();
							// <-----khaos- End Statistics Stuff
							
							client->m_lastPartAsked = 0xffff; // Reset current downloaded Chunk // Maella -Enhanced Chunk Selection- (based on jicxicmic)

							client->SendBlockRequests();
						}
					}
					else{
						if (!client->GetSentCancelTransfer()){
							Packet* packet = new Packet(OP_CANCELTRANSFER,0);
							theApp.uploadqueue->AddUpDataOverheadFileRequest(packet->size);
							client->socket->SendPacket(packet,true,true);
							client->SetSentCancelTransfer(1);
						}
						client->SetDownloadState((client->reqfile==NULL || client->reqfile->IsStopped()) ? DS_NONE : DS_ONQUEUE, CUpDownClient::DSR_CANCELLED); // Maella -Download Stop Reason- // [TPT]
					}
					break;
				}
				// [TPT] -  Maella -Code Improvement-
				case OP_REQUESTPARTS:{
					theApp.downloadqueue->AddDownDataOverheadFileRequest(size);
					client->CheckHandshakeFinished(OP_EDONKEYPROT, opcode);
					CSafeMemFile data((BYTE*)packet,size);
					Requested_Block_Struct* reqblock1 = new Requested_Block_Struct;
					Requested_Block_Struct* reqblock2 = new Requested_Block_Struct;
					Requested_Block_Struct* reqblock3 = new Requested_Block_Struct;
				data.Read(reqblock1->FileID, 16); // Small code Improvement
				md4cpy(reqblock2->FileID, reqblock1->FileID);
				md4cpy(reqblock3->FileID, reqblock1->FileID);
					data.Read(&reqblock1->StartOffset,4);
					data.Read(&reqblock2->StartOffset,4);
					data.Read(&reqblock3->StartOffset,4);
					data.Read(&reqblock1->EndOffset,4);
					data.Read(&reqblock2->EndOffset,4);
					data.Read(&reqblock3->EndOffset,4);

				// Maella -Close Backdoor-
				// Remark: There is a security leak that a leecher mod could exploit here.
				//         A client might send reqblock for another file than the one it 
				//         was granted to download. As long as the file ID in reqblock
				//         is the same in all reqblocks, it won't be rejected.  
				//         With this a client might be in a waiting queue with a high 
				//         priority but download block of a file set to a lower priority.
				if(md4cmp(reqblock1->FileID, client->GetUploadFileID()) != 0){
					// Maella -Filter verbose messages-
					if(theApp.glob_prefs->GetBlockUploadEndMsg() == false){
						AddPhoenixLogLine(false, "%s: Upload session ended due wrong requested FileID (client=%s, expected=%s, asked=%s)", 
											client->GetUserName(), client->GetClientSoftVer(), md4str(client->GetUploadFileID()), md4str(reqblock1->FileID));
					}
					// Maella end

					// Flag blocks to delete
					reqblock1->StartOffset = 0; reqblock1->EndOffset = 0; 
					reqblock2->StartOffset = 0; reqblock2->EndOffset = 0; 
					reqblock3->StartOffset = 0; reqblock3->EndOffset = 0; 

					// Send a 'out-of-part' indication
					Packet* packet = new Packet(OP_OUTOFPARTREQS, 0); // Missused?
					theApp.uploadqueue->AddUpDataOverheadOther(packet->size);
					client->socket->SendPacket(packet, true, true);

					// Try to transfer back the client to the waiting queue and update FileID
					// Remark: It's not necessary to check if this client must be added to the file as 
					//         a new source, because it should have already sent one time an OP_FILESTATUS request...
					//         It's important to lower the score of the client (see SetWaitStartTime), otherwise
					//         we will have a ping-pong effect if the client keeps asking two files with the same
					//         priority (e.g. granted for A but Asked B, then granted for B but asked A...)
					theApp.uploadqueue->RemoveFromUploadQueue(client, CUpDownClient::USR_DIFFERENT_FILE); // Maella -Upload Stop Reason-
					theApp.uploadqueue->AddClientToQueue(client, true); // Line to comment for hard punishement
					client->SetUploadFileID(reqblock1->FileID);
					client->SetWaitStartTime(); // Penality (soft punishement)

					// Ban Client (hard punishement)
					// theApp.ipfilter->AddBannedIPRange(client->GetIP(), client->GetIP(), 1, _T("Temporary"));
					// theApp.emuledlg->AddDebugLogLine(false, _T("Client %s with IP: %s has been added temporary to the filered list"), 
					// 								 client->GetUserName(), client->GetFullIP());
				}
				// Maella end

				if(reqblock1->StartOffset == reqblock1->EndOffset)
						delete reqblock1;
					else
						client->AddReqBlock(reqblock1);

				if(reqblock2->StartOffset == reqblock2->EndOffset)
						delete reqblock2;
					else
						client->AddReqBlock(reqblock2);

				if(reqblock3->StartOffset == reqblock3->EndOffset)
						delete reqblock3;
					else
						client->AddReqBlock(reqblock3);
					break;
				}
				// Maella end
				case OP_CANCELTRANSFER:{
					theApp.downloadqueue->AddDownDataOverheadFileRequest(size);
					client->CheckHandshakeFinished(OP_EDONKEYPROT, opcode);
					// [TPT] - Maella -Code Improvment-
					if(theApp.uploadqueue->RemoveFromUploadQueue(client, CUpDownClient::USR_CANCELLED) == true){ // Maella -Upload Stop Reason-
						// Maella -Filter verbose messages-
						if(theApp.glob_prefs->GetBlockUploadEndMsg() == false){
							AddDebugLogLine(false, "%s (client=%s): Upload session ended due canceled transfer.", client->GetUserName(), client->GetClientSoftVer());
						}
						//else
						//client->DoSnafu(SNAFU_OPCODE);//eWombat [SNAFU][TPT]
					client->SetUploadFileID(NULL);
					}
					// Maella end

					break;
				}
				case OP_END_OF_DOWNLOAD:{
					theApp.downloadqueue->AddDownDataOverheadFileRequest(size);
					client->CheckHandshakeFinished(OP_EDONKEYPROT, opcode);
					if (size>=16 && !md4cmp(client->GetUploadFileID(),packet)){
					// Maella -Code Improvment-
					if(theApp.uploadqueue->RemoveFromUploadQueue(client, CUpDownClient::USR_CANCELLED) == true){ // Maella -Upload Stop Reason-
						// [TPT] - Maella -Filter verbose messages-
						if(theApp.glob_prefs->GetBlockUploadEndMsg() == false){
							AddPhoenixLogLine(false, "%s: Upload session ended due ended transfer.", client->GetUserName());
						}
						// Maella end
						client->SetUploadFileID(NULL);
					}
					// Maella end
				}
					break;
				}
				case OP_HASHSETREQUEST:{
					theApp.downloadqueue->AddDownDataOverheadFileRequest(size);
					client->CheckHandshakeFinished(OP_EDONKEYPROT, opcode);
					if (size != 16)
						throw GetResString(IDS_ERR_WRONGHPACKAGESIZE);
					client->SendHashsetPacket(packet);
					break;
				}
				case OP_HASHSETANSWER:{
					theApp.downloadqueue->AddDownDataOverheadFileRequest(size);
					client->CheckHandshakeFinished(OP_EDONKEYPROT, opcode);
					client->ProcessHashSet(packet,size);
					break;
				}
				case OP_SENDINGPART:{
//					theApp.downloadqueue->AddDownDataOverheadOther(0, 24);
					client->CheckHandshakeFinished(OP_EDONKEYPROT, opcode);
					if (client->reqfile && !client->reqfile->IsStopped() && 
						(client->reqfile->GetStatus()==PS_READY || client->reqfile->GetStatus()==PS_EMPTY)){
						client->ProcessBlockPacket(packet,size);
						if (client->reqfile->IsStopped() || client->reqfile->GetStatus()==PS_PAUSED || client->reqfile->GetStatus()==PS_ERROR){
							if (!client->GetSentCancelTransfer()){
								Packet* packet = new Packet(OP_CANCELTRANSFER,0);
								theApp.uploadqueue->AddUpDataOverheadFileRequest(packet->size);
								client->socket->SendPacket(packet,true,true);
								client->SetSentCancelTransfer(1);
							}
								// [TPT]
							client->SetDownloadState(client->reqfile->IsStopped() ? DS_NONE : DS_ONQUEUE, CUpDownClient::DSR_CANCELLED); // Maella -Download Stop Reason-
						}
					}
					else{
						if (!client->GetSentCancelTransfer()){
							Packet* packet = new Packet(OP_CANCELTRANSFER,0);
							theApp.uploadqueue->AddUpDataOverheadFileRequest(packet->size);
							client->socket->SendPacket(packet,true,true);
							client->SetSentCancelTransfer(1);
						}
						// [TPT]
					client->SetDownloadState((client->reqfile==NULL || client->reqfile->IsStopped()) ? DS_NONE : DS_ONQUEUE, CUpDownClient::DSR_CANCELLED); // Maella -Download Stop Reason-
					}
					break;
				}
				case OP_OUTOFPARTREQS:{
					theApp.downloadqueue->AddDownDataOverheadFileRequest(size);
					client->CheckHandshakeFinished(OP_EDONKEYPROT, opcode);
					if (client->GetDownloadState() == DS_DOWNLOADING)
					// [TPT]
					client->SetDownloadState(DS_ONQUEUE, CUpDownClient::DSR_OUTOFPART); // Maella -Download Stop Reason-
					break;
				}
				case OP_SETREQFILEID:{
					theApp.downloadqueue->AddDownDataOverheadFileRequest(size);
					client->CheckHandshakeFinished(OP_EDONKEYPROT, opcode);

					// IP banned, no answer for this request
					if (client->IsBanned() ){
						break;						
					}

					// DbT:FileRequest
					if (size == 16){
						if (!client->GetWaitStartTime())
							client->SetWaitStartTime();

						CKnownFile* reqfile = theApp.sharedfiles->GetFileByID((uchar*)packet);
						if (!reqfile){
							// send file request no such file packet (0x48)
							Packet* replypacket = new Packet(OP_FILEREQANSNOFIL, 16);
							md4cpy(replypacket->pBuffer, packet);
							theApp.uploadqueue->AddUpDataOverheadFileRequest(replypacket->size);
							SendPacket(replypacket, true);
							break;
						}

						// if we are downloading this file, this could be a new source
						if (reqfile->IsPartFile())
							if( theApp.glob_prefs->GetMaxSourcePerFile() > ((CPartFile*)reqfile)->GetSourceCount() )
								theApp.downloadqueue->CheckAndAddKnownSource((CPartFile*)reqfile,client);
					            
						// check to see if this is a new file they are asking for
						if(md4cmp(client->GetUploadFileID(), packet) != 0)
							client->SetCommentDirty();

//						CKnownFile* clientreqfile = theApp.sharedfiles->GetFileByID(client->GetUploadFileID());
//						if( clientreqfile )
//							clientreqfile->SubQueuedCount();
//						reqfile->AddQueuedCount();

						//send filestatus
						client->SetUploadFileID((uchar*)packet);
//						md4cpy(client->reqfileid,packet);
						CSafeMemFile data(16+16);
						data.Write(reqfile->GetFileHash(),16);
						if (reqfile->IsPartFile())
							((CPartFile*)reqfile)->WritePartStatus(&data, client);	// [TPT] - SLUGFILLER: hideOS
						else if (!reqfile->HideOvershares(&data, client)){	// [TPT] - SLUGFILLER: hideOS
							uint32 null = 0;
							data.Write(&null,3);
						}
						Packet* packet = new Packet(&data);
						packet->opcode = OP_FILESTATUS;
						theApp.uploadqueue->AddUpDataOverheadFileRequest(packet->size);
						SendPacket(packet,true);
						break;
					}
					throw GetResString(IDS_ERR_WRONGPACKAGESIZE);
					break;
					// DbT:End
				}
				case OP_CHANGE_CLIENT_ID:{
					theApp.downloadqueue->AddDownDataOverheadOther(size);
					client->CheckHandshakeFinished(OP_EDONKEYPROT, opcode);
					CSafeMemFile data((BYTE*)packet, size);
					uint32 nNewUserID;
					data.Read(&nNewUserID, 4);
					uint32 nNewServerIP;
					data.Read(&nNewServerIP, 4);
					if (IsLowIDED2K(nNewUserID)){ // client changed server and gots a LowID
						CServer* pNewServer = theApp.serverlist->GetServerByIP(nNewServerIP);
						if (pNewServer != NULL){
							client->SetUserIDHybrid(nNewUserID); // update UserID only if we know the server
							client->SetServerIP(nNewServerIP);
							client->SetServerPort(pNewServer->GetPort());
						}
					}
					else if (nNewUserID == client->GetIP()){ // client changed server and gots a HighID(IP)
						client->SetUserIDHybrid(ntohl(nNewUserID));
						CServer* pNewServer = theApp.serverlist->GetServerByIP(nNewServerIP);
						if (pNewServer != NULL){
							client->SetServerIP(nNewServerIP);
							client->SetServerPort(pNewServer->GetPort());
						}
					}
					break;
				}
				case OP_CHANGE_SLOT:{
					theApp.downloadqueue->AddDownDataOverheadOther(size);
					client->CheckHandshakeFinished(OP_EDONKEYPROT, opcode);
					// sometimes sent by Hybrid
					break;
				}
				case OP_MESSAGE:{
					theApp.downloadqueue->AddDownDataOverheadOther(size);
					client->CheckHandshakeFinished(OP_EDONKEYPROT, opcode);
					if (size < 2)
						throw CString(_T("invalid message packet"));
					uint16 length;
					MEMCOPY(&length,packet,2);
					if ((UINT)(length+2) != size)
						throw CString(_T("invalid message packet"));

					AddDebugLogLine(true,"New Message from '%s' (IP:%s)",client->GetUserName(), client->GetFullIP());	
					//filter me?
					if ( (theApp.glob_prefs->MsgOnlyFriends() && !client->IsFriend()) ||
						 (theApp.glob_prefs->MsgOnlySecure() && client->GetUserName()==NULL) ) {
						if (!client->m_bMsgFiltered)
							AddDebugLogLine(false,"Filtered Message from '%s' (IP:%s)",client->GetUserName(), client->GetFullIP());
						client->m_bMsgFiltered=true;
						break;
					}
					char* message = new char[length+1];
					MEMCOPY(message,packet+2,length);
					message[length] = '\0';
					theApp.emuledlg->chatwnd->chatselector.ProcessMessage(client,message);
					delete[] message;
					break;
				}
				case OP_ASKSHAREDFILES:	{	// client wants to know what we have in share, let's see if we allow him to know that
					theApp.downloadqueue->AddDownDataOverheadOther(size);
					client->CheckHandshakeFinished(OP_EDONKEYPROT, opcode);
					
					// IP banned, no answer for this request
					if (client->IsBanned() ){
						break;						
					}

					CPtrList list;
					if (!client->IsSnafu() && !client->IsSUIFailed() && (theApp.glob_prefs->CanSeeShares() == 0 || (theApp.glob_prefs->CanSeeShares() == 1 && client->IsFriend()) ) )	// eWombat [SNAFU][TPT]
					{
						CCKey bufKey;
						CKnownFile* cur_file;
						for (POSITION pos = theApp.sharedfiles->m_Files_map.GetStartPosition();pos != 0;){
							theApp.sharedfiles->m_Files_map.GetNextAssoc(pos,bufKey,cur_file);
							// [TPT] - xMule_MOD: showSharePermissions
							// only show the files that you agree to show (because sometimes people would like to show
							// most of their files but hesitate to show a few ones (for some reasons :))
							if ( (cur_file->GetPermissions()==PERM_ALL)
								|| ((cur_file->GetPermissions()==PERM_FRIENDS) && client->IsFriend()) )
							list.AddTail((void*&)cur_file);
							// [TPT] - xMule_MOD: showSharePermissions
						}
						AddLogLine(true,GetResString(IDS_REQ_SHAREDFILES),client->GetUserName(),client->GetUserIDHybrid(),GetResString(IDS_ACCEPTED) );
					}
					else{
						AddLogLine(true,GetResString(IDS_REQ_SHAREDFILES),client->GetUserName(),client->GetUserIDHybrid(),GetResString(IDS_DENIED) );
					}

					// now create the memfile for the packet
					uint32 iTotalCount = list.GetCount();
					CSafeMemFile tempfile(80);
					tempfile.Write(&iTotalCount, 4);
					while (list.GetCount())
					{
						theApp.sharedfiles->CreateOfferedFilePacket((CKnownFile*)list.GetHead(), &tempfile, false);
						list.RemoveHead();
					}
					// create a packet and send it
					Packet* replypacket = new Packet(&tempfile);
					replypacket->opcode = OP_ASKSHAREDFILESANSWER;
					theApp.uploadqueue->AddUpDataOverheadOther(replypacket->size);
					SendPacket(replypacket, true, true);
					break;
				}
				case OP_ASKSHAREDFILESANSWER:{
					theApp.downloadqueue->AddDownDataOverheadOther(size);
					client->CheckHandshakeFinished(OP_EDONKEYPROT, opcode);
					client->ProcessSharedFileList(packet,size);
					break;
				}
                case OP_ASKSHAREDDIRS:{
                    theApp.downloadqueue->AddDownDataOverheadOther(size);
					client->CheckHandshakeFinished(OP_EDONKEYPROT, opcode);
                    ASSERT( size == 0 );
					
					// IP banned, no answer for this request
					if (client->IsBanned() ){
						break;						
					}

                    if (!client->IsSnafu() && !client->IsSUIFailed() && (theApp.glob_prefs->CanSeeShares()==0 || (theApp.glob_prefs->CanSeeShares()==1 && client->IsFriend()) )){
						AddLogLine(true,GetResString(IDS_SHAREDREQ1),client->GetUserName(),client->GetUserIDHybrid(),GetResString(IDS_ACCEPTED) );//eWombat [SNAFU][TPT]
 
						CString strDir,strTest;

						// [TPT] - SLUGFILLER: shareSubdir - enumerate according to shared files
						// Pichuei. This is used to process directories from files and be able to
						// get virtual and physical directories
						CStringList toSend;		// String list, because it's easier and faster
						CCKey bufKey;
						CKnownFile* cur_file;
						for (POSITION pos = theApp.sharedfiles->m_Files_map.GetStartPosition();pos != 0;){
							theApp.sharedfiles->m_Files_map.GetNextAssoc(pos,bufKey,cur_file);							
							// [TPT] - xMule_MOD: showSharePermissions - don't send dir names that are empty
							// due to file browse permissions
							if ( cur_file->GetPermissions() == PERM_NOONE 
								|| (cur_file->GetPermissions() == PERM_FRIENDS && !client->IsFriend()) )
								continue;
							// [TPT] - xMule_MOD: showSharePermissions
							CString path = cur_file->GetPath(true);	// itsonlyme: virtualDirs
							path.MakeLower();
							if (toSend.Find(path) == NULL)
								toSend.AddTail(path);
						}
						
						//build packet
                        CSafeMemFile tempfile(80);
						uint32 uDirs = toSend.GetCount();
                        tempfile.Write(&uDirs, 4);
						for (POSITION pos = toSend.GetHeadPosition();pos != 0;toSend.GetNext(pos))
                        {
							uint16 cnt = toSend.GetAt(pos).GetLength();
                            tempfile.Write(&cnt, 2);
							tempfile.Write((LPCTSTR)toSend.GetAt(pos), cnt);
                        }
						// [TPT] - SLUGFILLER: shareSubdir

						Packet* replypacket = new Packet(&tempfile);
                        replypacket->opcode = OP_ASKSHAREDDIRSANS;
                        theApp.uploadqueue->AddUpDataOverheadOther(replypacket->size);
                        SendPacket(replypacket, true, true);
					}
					else{
						AddLogLine(true,GetResString(IDS_SHAREDREQ1),client->GetUserName(),client->GetUserIDHybrid(),GetResString(IDS_DENIED) );
                        Packet* replypacket = new Packet(OP_ASKSHAREDDENIEDANS, 0);
						theApp.uploadqueue->AddUpDataOverheadOther(replypacket->size);
                        SendPacket(replypacket, true, true);
                    }
                    break;
                }
                case OP_ASKSHAREDFILESDIR:{
                    theApp.downloadqueue->AddDownDataOverheadOther(size);
					client->CheckHandshakeFinished(OP_EDONKEYPROT, opcode);
					
					// IP banned, no answer for this request
					if (client->IsBanned() ){
						break;						
					}

                    CSafeMemFile data((uchar*)packet, size);
                    uint16 cnt;
                    data.Read(&cnt, 2);
                    CString strReqDir;
                    data.Read(strReqDir.GetBuffer(cnt), cnt);
                    strReqDir.ReleaseBuffer(cnt);
                    PathRemoveBackslash(strReqDir.GetBuffer());
                    strReqDir.ReleaseBuffer();
 
                    if (!client->IsSnafu() && !client->IsSUIFailed() && (theApp.glob_prefs->CanSeeShares()==0 || (theApp.glob_prefs->CanSeeShares()==1 && client->IsFriend()))){
						AddLogLine(true,GetResString(IDS_SHAREDREQ2),client->GetUserName(),client->GetUserIDHybrid(),strReqDir,GetResString(IDS_ACCEPTED) );//eWombat [SNAFU][TPT]
                        ASSERT( data.GetPosition() == data.GetLength() );
 
                        CTypedPtrList<CPtrList, CKnownFile*> list;
                        for (POSITION pos = theApp.sharedfiles->m_Files_map.GetStartPosition();pos != 0;){
                            CCKey bufKey;
                            CKnownFile* cur_file;
                            theApp.sharedfiles->m_Files_map.GetNextAssoc(pos, bufKey, cur_file);
							CString strSharedFileDir(cur_file->GetPath(true));	// itsonlyme: virtualDirs
							PathRemoveBackslash(strSharedFileDir.GetBuffer());
							strSharedFileDir.ReleaseBuffer();
							// [TPT] - xMule_MOD: showSharePermissions
							// only show the files that you agree to show (because sometimes people would like to show
							// most of their files but hesitate to show a few ones (for some reasons :))
							if ( strReqDir.CompareNoCase(strSharedFileDir) == 0 &&
								( (cur_file->GetPermissions()==PERM_ALL)
								|| ((cur_file->GetPermissions()==PERM_FRIENDS) && client->IsFriend()) ) )
                                list.AddTail(cur_file);
							// [TPT] - xMule_MOD: showSharePermissions
                        }
 
						// Currently we are sending each shared directory, even if it does not contain any files.
						// Because of this we also have to send an empty shared files list..
						/*if (list.GetCount())*/{
                            CSafeMemFile tempfile(80);
                            uint16 cnt = strReqDir.GetLength();
                            tempfile.Write(&cnt, 2);
                            tempfile.Write((LPCTSTR)strReqDir, cnt);
                            uint32 uFiles = list.GetCount();
                            tempfile.Write(&uFiles, 4);
                            while (list.GetCount()){
                                theApp.sharedfiles->CreateOfferedFilePacket(list.GetHead(), &tempfile, false);
                                list.RemoveHead();
                            }
 
                            Packet* replypacket = new Packet(&tempfile);
                            replypacket->opcode = OP_ASKSHAREDFILESDIRANS;
                            theApp.uploadqueue->AddUpDataOverheadOther(replypacket->size);
                            SendPacket(replypacket, true, true);
                        }
                    }
                    else{
						AddLogLine(true,GetResString(IDS_SHAREDREQ2),client->GetUserName(),client->GetUserIDHybrid(),strReqDir,GetResString(IDS_DENIED) );
                        Packet* replypacket = new Packet(OP_ASKSHAREDDENIEDANS, 0);
                        theApp.uploadqueue->AddUpDataOverheadOther(replypacket->size);
                        SendPacket(replypacket, true, true);
                    }
                    break;
                }
                case OP_ASKSHAREDDIRSANS:{
                    theApp.downloadqueue->AddDownDataOverheadOther(size);
					client->CheckHandshakeFinished(OP_EDONKEYPROT, opcode);
                    client->ProcessSharedDirsList(packet,size);	// [TPT] - itsonlyme: viewSharedFiles
                    break;
                }
                case OP_ASKSHAREDFILESDIRANS:{
                    theApp.downloadqueue->AddDownDataOverheadOther(size);
					client->CheckHandshakeFinished(OP_EDONKEYPROT, opcode);
                    CSafeMemFile data((uchar*)packet, size);
                    uint16 cnt;
                    data.Read(&cnt, 2);
                    CString strDir;
                    data.Read(strDir.GetBuffer(cnt), cnt);
                    strDir.ReleaseBuffer(cnt);
					PathRemoveBackslash(strDir.GetBuffer());
					strDir.ReleaseBuffer();
                    if (client->GetFileListRequested() > 0){
						AddDebugLogLine(true,GetResString(IDS_SHAREDINFO1),client->GetUserName(),client->GetUserIDHybrid(),strDir);	// [TPT] - itsonlyme: viewSharedFiles
						client->ProcessSharedFileList(packet + data.GetPosition(), size - data.GetPosition(), strDir);
						if (client->GetFileListRequested() == 0)
							AddDebugLogLine(true,GetResString(IDS_SHAREDINFO2),client->GetUserName(),client->GetUserIDHybrid());	// [TPT] - itsonlyme: viewSharedFiles
                    }
					else
						AddDebugLogLine(true,GetResString(IDS_SHAREDANSW3),client->GetUserName(),client->GetUserIDHybrid(),strDir);	// [TPT] - itsonlyme: viewSharedFiles
                    break;
                }
                case OP_ASKSHAREDDENIEDANS:{
                    theApp.downloadqueue->AddDownDataOverheadOther(size);
					client->CheckHandshakeFinished(OP_EDONKEYPROT, opcode);
                    ASSERT( size == 0 );
					AddDebugLogLine(true,GetResString(IDS_SHAREDREQDENIED),client->GetUserName(),client->GetUserIDHybrid());	// [TPT] - itsonlyme: viewSharedFiles
					client->SetDeniesShare();	// [TPT] - itsonlyme: viewSharedFiles
					client->SetFileListRequested(0);
                    break;
                }

				default:
					theApp.downloadqueue->AddDownDataOverheadOther(size);
				// [TPT]
				PacketToDebugLogLine(false, packet, size, opcode); 
					break;
			}
		}
	// Maella -Catch incoming packets exception (wrong size)-
	catch(CString error){
		if (client != NULL){						
			AddDebugLogLine(false, _T("%s - while processing eDonkey packet: opcode=%s  size=%u; %s"), error, DbgGetDonkeyClientTCPOpcode(opcode), size, DbgGetClientInfo());
			
			client->SetDownloadState(DS_ERROR, CUpDownClient::DSR_EXCEPTION); // Maella -Download Stop Reason-
		}
		else {
			SOCKADDR_IN sockAddr;
			MEMSET(&sockAddr, 0, sizeof(sockAddr));
			uint32 nSockAddrLen = sizeof(sockAddr);
			GetPeerName((SOCKADDR*)&sockAddr,(int*)&nSockAddrLen);

			AddDebugLogLine(false, _T("Client with IP=%s caused an error or did something bad: %s. Disconnecting client!"), inet_ntoa(sockAddr.sin_addr), error.GetBuffer());
		}
		PacketToDebugLogLine(true, packet, size, opcode);	
		Disconnect();
		return false;
	}
	catch(CFileException* pFileException){
		pFileException->Delete();
		AddDebugLogLine(true, _T("A client has caused an exception. Disconnecting client!"));		
		PacketToDebugLogLine(true, packet, size, opcode);
		// Maella -Temporary Ban IP of clients causing an error-
		if (client != NULL){ 			
			theApp.ipfilter->AddBannedIPRange(client->GetIP(), client->GetIP(), 1, _T("Temporary"));
			AddDebugLogLine(false, _T("Client %s with IP: %s has been added temporary to the filered list"), client->GetUserName(), client->GetFullIP());			
			client->SetDownloadState(DS_ERROR, CUpDownClient::DSR_EXCEPTION); // Maella -Download Stop Reason-
		}
		else {
			SOCKADDR_IN sockAddr;
			MEMSET(&sockAddr, 0, sizeof(sockAddr));
			uint32 nSockAddrLen = sizeof(sockAddr);
			GetPeerName((SOCKADDR*)&sockAddr,(int*)&nSockAddrLen);

			theApp.ipfilter->AddBannedIPRange(sockAddr.sin_addr.S_un.S_addr, sockAddr.sin_addr.S_un.S_addr, 1, _T("Temporary"));
			AddDebugLogLine(false, _T("IP: %s has been added temporary to the filered list"), inet_ntoa(sockAddr.sin_addr));			
	}
	// Maella end
		Disconnect();
		return false;
	}
	catch(...){
		AddDebugLogLine(true, _T("A client has caused an undefined exception. Disconnecting client!"));		
		PacketToDebugLogLine(true, packet, size, opcode);
		// Maella -Temporary Ban IP of clients causing an error-
		if (client != NULL){ 			
			theApp.ipfilter->AddBannedIPRange(client->GetIP(), client->GetIP(), 1, _T("Temporary"));
			AddDebugLogLine(false, _T("Client %s with IP: %s has been added temporary to the filered list"), client->GetUserName(), client->GetFullIP());			
			client->SetDownloadState(DS_ERROR, CUpDownClient::DSR_EXCEPTION); // Maella -Download Stop Reason-
		}
		else {
			SOCKADDR_IN sockAddr;
			MEMSET(&sockAddr, 0, sizeof(sockAddr));
			uint32 nSockAddrLen = sizeof(sockAddr);
			GetPeerName((SOCKADDR*)&sockAddr,(int*)&nSockAddrLen);

			theApp.ipfilter->AddBannedIPRange(sockAddr.sin_addr.S_un.S_addr, sockAddr.sin_addr.S_un.S_addr, 1, _T("Temporary"));
			AddDebugLogLine(false, _T("IP: %s has been added temporary to the filered list"), inet_ntoa(sockAddr.sin_addr));			
		}		
		// Maella end
		Disconnect();
		return false;
	}
	// Maella end
	return true;
}

bool CClientReqSocket::ProcessExtPacket(char* packet, uint32 size, UINT opcode, UINT uRawSize){
	try{
		//try{ // [TPT]
			if (!client){
				theApp.downloadqueue->AddDownDataOverheadOther(uRawSize);
				throw GetResString(IDS_ERR_UNKNOWNCLIENTACTION);
			}
			if (_iDbgHeap >= 2)
				ASSERT_VALID(client);
			switch(opcode){
				case OP_EMULEINFO:{
					theApp.downloadqueue->AddDownDataOverheadOther(uRawSize);
					client->ProcessMuleInfoPacket(packet,size);

					// start secure identification, if
					//  - we have received eD2K and eMule info (old eMule)
					if (client->GetInfoPacketsReceived() == IP_BOTH)
						client->InfoPacketsReceived();

					client->SendMuleInfoPacket(true);
					break;
				}
				case OP_EMULEINFOANSWER:{
					theApp.downloadqueue->AddDownDataOverheadOther(uRawSize);
					client->ProcessMuleInfoPacket(packet,size);

					// start secure identification, if
					//  - we have received eD2K and eMule info (old eMule)
					if (client->GetInfoPacketsReceived() == IP_BOTH)
						client->InfoPacketsReceived();
					break;
				}
				case OP_SECIDENTSTATE:{
					theApp.downloadqueue->AddDownDataOverheadOther(uRawSize);
					client->ProcessSecIdentStatePacket((uchar*)packet,size);
					if (client->GetSecureIdentState() == IS_SIGNATURENEEDED)
						client->SendSignaturePacket();
					else if (client->GetSecureIdentState() == IS_KEYANDSIGNEEDED){
						client->SendPublicKeyPacket();
						client->SendSignaturePacket();
					}
					break;
				}
				case OP_PUBLICKEY:{
					theApp.downloadqueue->AddDownDataOverheadOther(uRawSize);
					if (client->IsBanned() ){
						break;						
					}
					client->ProcessPublicKeyPacket((uchar*)packet,size);
					break;
				}
  				case OP_SIGNATURE:{
					theApp.downloadqueue->AddDownDataOverheadOther(uRawSize);
					client->ProcessSignaturePacket((uchar*)packet,size);
					break;
				}
				case OP_COMPRESSEDPART:{
	//				theApp.downloadqueue->AddDownDataOverheadOther(24);
					client->CheckHandshakeFinished(OP_EMULEPROT, opcode);
					if (client->reqfile && !client->reqfile->IsStopped() && 
						(client->reqfile->GetStatus()==PS_READY || client->reqfile->GetStatus()==PS_EMPTY)){
						client->ProcessBlockPacket(packet,size,true);
						if (client->reqfile->IsStopped() || client->reqfile->GetStatus()==PS_PAUSED || client->reqfile->GetStatus()==PS_ERROR){
							if (!client->GetSentCancelTransfer()){
								Packet* packet = new Packet(OP_CANCELTRANSFER,0);
								theApp.uploadqueue->AddUpDataOverheadFileRequest(packet->size);
								client->socket->SendPacket(packet,true,true);
								client->SetSentCancelTransfer(1);
							}
							// [TPT]
						client->SetDownloadState(client->reqfile->IsStopped() ? DS_NONE : DS_ONQUEUE, CUpDownClient::DSR_CANCELLED); // Maella -Download Stop Reason-
						}
					}
					else{
						if (!client->GetSentCancelTransfer()){
							Packet* packet = new Packet(OP_CANCELTRANSFER,0);
							theApp.uploadqueue->AddUpDataOverheadFileRequest(packet->size);
							client->socket->SendPacket(packet,true,true);
							client->SetSentCancelTransfer(1);
						}
							// [TPT]
					client->SetDownloadState((client->reqfile==NULL || client->reqfile->IsStopped()) ? DS_NONE : DS_ONQUEUE, CUpDownClient::DSR_CANCELLED); // Maella -Download Stop Reason-
					}
					break;
				}
				case OP_QUEUERANKING:{
					theApp.downloadqueue->AddDownDataOverheadFileRequest(uRawSize);
					client->CheckHandshakeFinished(OP_EMULEPROT, opcode);
					if (size != 12)
						throw GetResString(IDS_ERR_BADSIZE);
					uint16 newrank;
					MEMCOPY(&newrank,packet+0,2);
					client->SetRemoteQueueFull(false);
					client->SetRemoteQueueRank(newrank);
					break;
				}
 				case OP_REQUESTSOURCES:{
					theApp.downloadqueue->AddDownDataOverheadSourceExchange(uRawSize);
					client->CheckHandshakeFinished(OP_EMULEPROT, opcode);
					//eWombat [SNAFU][TPT]
					if (client->IsSnafu() || client->IsNotSUI())
						break;
					//eWombat [SNAFU][TPT]
					if (client->GetSourceExchangeVersion() > 1){
						if(size != 16)
							throw GetResString(IDS_ERR_BADSIZE);
				// [TPT]
				// Maella -Enable/Disable source exchange in preference- (Tarod)
				if (theApp.glob_prefs->GetDisabledXS())
					break;
				// Maella end
		
						//first check shared file list, then download list
						CKnownFile* file = theApp.sharedfiles->GetFileByID((uchar*)packet);
						if(!file)
							file = theApp.downloadqueue->GetFileByID((uchar*)packet);
						if(file) {
							DWORD dwTimePassed = ::GetTickCount() - client->GetLastSrcReqTime() + CONNECTION_LATENCY;
							bool bNeverAskedBefore = client->GetLastSrcReqTime() == 0;
		
							if( 
								//if not complete and file is rare, allow once every 10 minutes
								(    file->IsPartFile()
								  && (bNeverAskedBefore || dwTimePassed > SOURCECLIENTREASK)
								  && ((CPartFile*)file)->GetSourceCount() <= RARE_FILE * 2
								) ||
								//OR if file is not rare or if file is complete, allow every 90 minutes
								( (bNeverAskedBefore || dwTimePassed > SOURCECLIENTREASK * MINCOMMONPENALTY) )
							) {
								client->SetLastSrcReqTime();
								Packet* tosend = file->CreateSrcInfoPacket(client);
								if(tosend){
									if ( theApp.glob_prefs->GetDebugSourceExchange() )
										AddDebugLogLine( false, "RCV:Source Request User(%s) File(%s)", client->GetUserName(), file->GetFileName() );
									theApp.uploadqueue->AddUpDataOverheadSourceExchange(tosend->size);
									SendPacket(tosend, true, true);
								}
							}
						}
					}
					break;
				}
 				case OP_ANSWERSOURCES:{
					theApp.downloadqueue->AddDownDataOverheadSourceExchange(uRawSize);
					client->CheckHandshakeFinished(OP_EMULEPROT, opcode);
					CSafeMemFile data((BYTE*)packet,size);
					uchar hash[16];
					data.Read(hash,16);
					CKnownFile* file = theApp.downloadqueue->GetFileByID((uchar*)packet);
					if(file){
						if (file->IsPartFile()){
							//set the client's answer time
							client->SetLastSrcAnswerTime();
							//and set the file's last answer time
							((CPartFile*)file)->SetLastAnsweredTime();
							((CPartFile*)file)->AddClientSources(&data, client->GetSourceExchangeVersion());
						}
					}
					break;
				}
				case OP_FILEDESC:{
					theApp.downloadqueue->AddDownDataOverheadFileRequest(uRawSize);
					client->CheckHandshakeFinished(OP_EMULEPROT, opcode);
					client->ProcessMuleCommentPacket(packet,size);
					break;
				}
				case OP_REQUESTPREVIEW:{
					theApp.downloadqueue->AddDownDataOverheadOther(uRawSize);
					client->CheckHandshakeFinished(OP_EMULEPROT, opcode);
					// IP banned, no answer for this request
					if (client->IsBanned() ){
						break;						
					}

					if (theApp.glob_prefs->CanSeeShares() == 0 					// everybody
						|| (theApp.glob_prefs->CanSeeShares() == 1 && client->IsFriend() )// friend
						&& theApp.glob_prefs->IsPreviewEnabled() )	
					{
						client->ProcessPreviewReq(packet,size);	
						AddDebugLogLine(true,"Client '%s' (%s) requested Preview - accepted",client->GetUserName(),client->GetFullIP());
					}
					else{
						// we don't send any answer here, because the client should know that he was not allowed to ask
						AddDebugLogLine(true,"Client '%s' (%s) requested Preview - denied",client->GetUserName(),client->GetFullIP());
					}
					break;
				}
				case OP_PREVIEWANSWER:{
					theApp.downloadqueue->AddDownDataOverheadOther(uRawSize);
					client->CheckHandshakeFinished(OP_EMULEPROT, opcode);
					client->ProcessPreviewAnswer(packet, size);
					break;
				}
				default:
					theApp.downloadqueue->AddDownDataOverheadOther(uRawSize);
					PacketToDebugLogLine("eMule", packet, size, opcode);
					break;
			}
		}
	// [TPT]
		// Maella -Catch incoming packets exception (wrong size)-
	catch(CString error){			
		if (client != NULL){			
			AddDebugLogLine(false, _T("%s - while processing eMule packet: opcode=%s  size=%u; %s"), error, DbgGetMuleClientTCPOpcode(opcode), size, DbgGetClientInfo());
			client->SetDownloadState(DS_ERROR, CUpDownClient::DSR_EXCEPTION); // Maella -Download Stop Reason-
		}
		else {
			SOCKADDR_IN sockAddr;
			MEMSET(&sockAddr, 0, sizeof(sockAddr));
			uint32 nSockAddrLen = sizeof(sockAddr);
			GetPeerName((SOCKADDR*)&sockAddr,(int*)&nSockAddrLen);

			AddDebugLogLine(false, _T("Client with IP=%s caused an error or did something bad: %s. Disconnecting client!"), inet_ntoa(sockAddr.sin_addr), error.GetBuffer());
		}
		PacketToDebugLogLine(true, packet, size, opcode);	
		Disconnect();
		return false;
		}
	catch(CFileException* pFileException){
		pFileException->Delete();
		AddDebugLogLine(true, _T("A client has caused an exception. Disconnecting client!"));		
		PacketToDebugLogLine(true, packet, size, opcode);
		// Maella -Temporary Ban IP of clients causing an error-
		if (client != NULL){ 			
			theApp.ipfilter->AddBannedIPRange(client->GetIP(), client->GetIP(), 1, _T("Temporary"));
			AddDebugLogLine(false, _T("Client %s with IP: %s has been added temporary to the filered list"), client->GetUserName(), client->GetFullIP());			
			client->SetDownloadState(DS_ERROR, CUpDownClient::DSR_EXCEPTION); // Maella -Download Stop Reason-
		}
		else {
			SOCKADDR_IN sockAddr;
			MEMSET(&sockAddr, 0, sizeof(sockAddr));
			uint32 nSockAddrLen = sizeof(sockAddr);
			GetPeerName((SOCKADDR*)&sockAddr,(int*)&nSockAddrLen);

			theApp.ipfilter->AddBannedIPRange(sockAddr.sin_addr.S_un.S_addr, sockAddr.sin_addr.S_un.S_addr, 1, _T("Temporary"));
			AddDebugLogLine(false, _T("IP: %s has been added temporary to the filered list"), inet_ntoa(sockAddr.sin_addr));			
		}		
		// Maella end
		Disconnect();
		return false;
	}
	catch(...){
		AddDebugLogLine(true, _T("A client has caused an undefined exception. Disconnecting client!"));		
		PacketToDebugLogLine(true, packet, size, opcode);
		// Maella -Temporary Ban IP of clients causing an error-
		if (client != NULL){ 			
			theApp.ipfilter->AddBannedIPRange(client->GetIP(), client->GetIP(), 1, _T("Temporary"));
			AddDebugLogLine(false, _T("Client %s with IP: %s has been added temporary to the filered list"), client->GetUserName(), client->GetFullIP());			
			client->SetDownloadState(DS_ERROR, CUpDownClient::DSR_EXCEPTION); // Maella -Download Stop Reason-
	}
		else {
			SOCKADDR_IN sockAddr;
			MEMSET(&sockAddr, 0, sizeof(sockAddr));
			uint32 nSockAddrLen = sizeof(sockAddr);
			GetPeerName((SOCKADDR*)&sockAddr,(int*)&nSockAddrLen);

			theApp.ipfilter->AddBannedIPRange(sockAddr.sin_addr.S_un.S_addr, sockAddr.sin_addr.S_un.S_addr, 1, _T("Temporary"));
			AddDebugLogLine(false, _T("IP: %s has been added temporary to the filered list"), inet_ntoa(sockAddr.sin_addr));			
		}		
		// Maella end
		Disconnect();
		return false;
	}
	// Maella end
	return true;
}

// [TPT]
// Maella -Dump information of unknown packet in debug tab-
void CClientReqSocket::PacketToDebugLogLine(bool isOpcodeKnown, const char* packet, uint32 size, UINT opcode)  {
	
	// Dump Packet information	
	CString buffer; 
	if(isOpcodeKnown == false)
		buffer.Format(_T("Packet with unknown opcode: 0x%x, size=%d"), opcode, size);
	else
		buffer.Format(_T("Invalid packet with opcode: 0x%x, size=%d"), opcode, size);
	
	// Client information
	if(client != NULL){
		CString uploadString = _T("unknown");
		switch (client->GetUploadState()){
		case US_ONUPLOADQUEUE:
			uploadString = GetResString(IDS_ONQUEUE);
			break;
		case US_PENDING:
			uploadString = GetResString(IDS_CL_PENDING);
			break;
		case US_LOWTOLOWIP:
			uploadString = GetResString(IDS_CL_LOW2LOW);
			break;
		case US_BANNED:
			uploadString = GetResString(IDS_BANNED);
			break;
		case US_ERROR:
			uploadString = GetResString(IDS_ERROR);
			break;
		case US_CONNECTING:
			uploadString = GetResString(IDS_CONNECTING);
			break;
		case US_WAITCALLBACK:
			uploadString = GetResString(IDS_CONNVIASERVER);
			break;
		case US_UPLOADING:
			uploadString = GetResString(IDS_TRANSFERRING);
			break;
		case US_NONE:
			uploadString = _T("none");
			break;
		}

		CString downloadString = _T("unknown");
		switch (client->GetDownloadState()) {
		case DS_CONNECTING:
			downloadString = GetResString(IDS_CONNECTING);
			break;
		case DS_CONNECTED:
			downloadString = GetResString(IDS_ASKING);
			break;
		case DS_WAITCALLBACK:
			downloadString = GetResString(IDS_CONNVIASERVER);
			break;
		case DS_ONQUEUE:
			if( client->IsRemoteQueueFull() )
				downloadString = GetResString(IDS_QUEUEFULL);
			else
				downloadString = GetResString(IDS_ONQUEUE);
			break;
		case DS_DOWNLOADING:
			downloadString = GetResString(IDS_TRANSFERRING);
			break;
		case DS_REQHASHSET:
			downloadString = GetResString(IDS_RECHASHSET);
			break;
		case DS_NONEEDEDPARTS:
			downloadString = GetResString(IDS_NONEEDEDPARTS);
			break;
		case DS_LOWTOLOWIP:
			downloadString = GetResString(IDS_NOCONNECTLOW2LOW);
			break;
		case DS_TOOMANYCONNS:
			downloadString = GetResString(IDS_TOOMANYCONNS);
			break;
		case DS_NONE:
			downloadString = _T("none");
			break;
		}

		buffer += _T(", client=") + client->GetClientSoftVer();
		buffer += _T(", up. state=") + uploadString;
		buffer += _T(", down. state=") + downloadString;
	}
	// Hex packet dump
	buffer += _T(", data=[");	
	for (uint32 i = 0; i < size && i < 50; i++){
		if (i > 0)
			buffer += _T(' ');
		TCHAR temp[33];
		buffer += itoa((unsigned char)packet[i], temp, 16);		
	}
	buffer += (i == size) ? _T("]") : _T("..]");
	DbgAppendClientInfo(buffer);
	AddDebugLogLine(false, _T("%s"), buffer);
}
// Maella end

CString CClientReqSocket::DbgGetClientInfo()
{
	CString str;
	SOCKADDR_IN sockAddr = {0};
	int nSockAddrLen = sizeof(sockAddr);
	GetPeerName((SOCKADDR*)&sockAddr, &nSockAddrLen);
	if (sockAddr.sin_addr.S_un.S_addr != 0 && (client == NULL || sockAddr.sin_addr.S_un.S_addr != client->GetIP()))
		str.AppendFormat(_T("IP=%s"), inet_ntoa(sockAddr.sin_addr));
	if (client){
		if (!str.IsEmpty())
			str += _T("; ");
		str += _T("client=") + client->DbgGetClientInfo();
	}
	return str;
}

void CClientReqSocket::DbgAppendClientInfo(CString& str)
{
	CString strClientInfo(DbgGetClientInfo());
	if (!strClientInfo.IsEmpty()){
		if (!str.IsEmpty())
			str += _T("; ");
		str += strClientInfo;
	}
}

void CClientReqSocket::OnInit(){
	//uint8 tv = 1;
	//SetSockOpt(SO_DONTLINGER,&tv,sizeof(BOOL));
}

void CClientReqSocket::OnConnect(int nErrorCode)
{
	CEMSocket::OnConnect(nErrorCode);
	if (nErrorCode)
	{
		CString strTCPError = GetErrorMessage(nErrorCode, 1);

		// WSAECONNREFUSED (10061) Connection refused.
		// No connection could be made because the target machine actively refused it.
		// This usually results from trying to connect to a service that is inactive on the foreign host
		// that is, one with no server application running.
		//
		// WSAETIMEDOUT (10060) Connection timed out.
		// A connection attempt failed because the connected party did not properly respond after a
		// period of time, or the established connection failed because the connected host has failed to respond.
		if (nErrorCode != WSAECONNREFUSED && nErrorCode != WSAETIMEDOUT)
			AddDebugLogLine(false, _T("Client TCP socket error (OnConnect): %s; %s"), strTCPError, DbgGetClientInfo());
	}
}

void CClientReqSocket::OnSend(int nErrorCode){
	ResetTimeOutTimer();
	CEMSocket::OnSend(nErrorCode);
}

void CClientReqSocket::OnError(int nErrorCode)
{
	CString strTCPError;
	if (nErrorCode == ERR_WRONGHEADER)
		strTCPError = _T("Wrong header");
	else if (nErrorCode == ERR_TOOBIG)
		strTCPError = _T("Too much data sent");
	else
		strTCPError = GetErrorMessage(nErrorCode);

	AddDebugLogLine(false, _T("Client TCP socket error: %s; %s"), strTCPError, DbgGetClientInfo());

	Disconnect();
}

bool CClientReqSocket::PacketReceivedCppEx(Packet* packet)
{
	bool bResult;
	UINT uRawSize = packet->size;
	switch (packet->prot){
		case OP_EDONKEYPROT:
			bResult = ProcessPacket(packet->pBuffer,packet->size,packet->opcode);
			break;
		case OP_PACKEDPROT:
			if (!packet->UnPackPacket()){
				AddDebugLogLine(false, _T("Failed to decompress client TCP packet; %s; %s"), DbgGetClientTCPPacket(packet->prot, packet->opcode, packet->size), DbgGetClientInfo());
				bResult = false;
				break;
			}
		case OP_EMULEPROT:
			bResult = ProcessExtPacket(packet->pBuffer, packet->size, packet->opcode, uRawSize);
			break;
		default:{
			theApp.downloadqueue->AddDownDataOverheadOther(uRawSize);
			AddDebugLogLine(false, _T("Received unknown client TCP packet; %s; %s"), DbgGetClientTCPPacket(packet->prot, packet->opcode, packet->size), DbgGetClientInfo());

			if (client)
				client->SetDownloadState(DS_ERROR);
			Disconnect();
			bResult = false;
		}
	}
	return bResult;
}

#ifdef USE_CLIENT_TCP_CATCH_ALL_HANDLER
int FilterSE(DWORD dwExCode, LPEXCEPTION_POINTERS pExPtrs, CClientReqSocket* reqsock, Packet* packet)
{
	CString strExError;
	if (pExPtrs){
		const EXCEPTION_RECORD* er = pExPtrs->ExceptionRecord;
		strExError.Format(_T("Error: Unknown exception %08x in CClientReqSocket::PacketReceived at 0x%08x"), er->ExceptionCode, er->ExceptionAddress);
	}
	else
		strExError.Format(_T("Error: Unknown exception %08x in CClientReqSocket::PacketReceived"), dwExCode);

	// we already had an unknown exception, better be prepared for dealing with invalid data -> use another exception handler
	try{
		CString strError = strExError;
		strError.AppendFormat(_T("; %s"), DbgGetClientTCPPacket(packet?packet->prot:0, packet?packet->opcode:0, packet?packet->size:0));
		reqsock->DbgAppendClientInfo(strError);
		CemuleApp::AddDebugLogLine(false, _T("%s"), strError);
	}
	catch(...){
		ASSERT(0);
		CemuleApp::AddDebugLogLine(false, _T("%s"), strExError);
	}
	
	// this searches the next exception handler -> catch(...) in 'CAsyncSocketExHelperWindow::WindowProc'
	// as long as I do not know where and why we are crashing, I prefere to have it handled that way which
	// worked fine in 28a/b.
	//
	// 03-Jn-2004 [bc]: Returning the execution to the catch-all handler in 'CAsyncSocketExHelperWindow::WindowProc'
	// can make things even worse, because in some situations, the socket will continue fireing received events. And
	// because the processed packet (which has thrown the exception) was not removed from the EMSocket buffers, it would
	// be processed again and again.
	//return EXCEPTION_CONTINUE_SEARCH;

	// this would continue the program "as usual" -> return execution to the '__except' handler
	return EXCEPTION_EXECUTE_HANDLER;
}
#endif

bool CClientReqSocket::PacketReceived(Packet* packet)
{
	bool bResult;
#ifdef USE_CLIENT_TCP_CATCH_ALL_HANDLER
	bool bUnknownException = false;
	// this function is only here to get a chance of determining the crash address via SEH
	__try{
		bResult = PacketReceivedCppEx(packet);
	}
	__except(FilterSE(GetExceptionCode(), GetExceptionInformation(), this, packet)){
		bUnknownException = true;
	}
	if (bUnknownException){
		 if (client)
			 client->SetDownloadState(DS_ERROR);
		 Disconnect();
		 bResult = false;
	}
#else
	bResult = PacketReceivedCppEx(packet);
#endif
	return bResult;
}

void CClientReqSocket::OnReceive(int nErrorCode){
	ResetTimeOutTimer();
	CEMSocket::OnReceive(nErrorCode);
}

bool CClientReqSocket::Create(){
	theApp.listensocket->AddConnection();
	BOOL result = CAsyncSocketEx::Create(0,SOCK_STREAM,FD_WRITE|FD_READ|FD_CLOSE|FD_CONNECT); // deadlake PROXYSUPPORT - changed to AsyncSocketEx
	OnInit();
	return result;
}

// Pichuei
/*SocketSentBytes CClientReqSocket::Send(uint32 maxNumberOfBytesToSend, bool onlyAllowedToSendControlPacket) {
    SocketSentBytes returnStatus = CEMSocket::Send(maxNumberOfBytesToSend, onlyAllowedToSendControlPacket);

    if(returnStatus.success && (returnStatus.sentBytesControlPackets > 0 || returnStatus.sentBytesStandardPackets > 0)) {
        ResetTimeOutTimer();
    }

    return returnStatus;
}*/

// CListenSocket
// CListenSocket member functions
CListenSocket::CListenSocket(CPreferences* in_prefs)
{
	bListening = false;
	app_prefs = in_prefs;
	opensockets = 0;
	maxconnectionreached = 0;
	m_OpenSocketsInterval = 0;
	m_nPeningConnections = 0;
	MEMSET(m_ConnectionStates, 0, sizeof m_ConnectionStates);
	peakconnections = 0;
	totalconnectionchecks = 0;
	averageconnections = 0.0;
	activeconnections = 0;
}

CListenSocket::~CListenSocket(){
	Close();
	KillAllSockets();
}

bool CListenSocket::StartListening(){
	bListening = true;
	return (this->Create(app_prefs->GetPort(),SOCK_STREAM,FD_ACCEPT) && this->Listen());
}

void CListenSocket::ReStartListening(){
	bListening = true;
	if (m_nPeningConnections){
		m_nPeningConnections--;
		OnAccept(0);
	}
}

void CListenSocket::StopListening(){
	bListening = false;
	maxconnectionreached++;
}

void CListenSocket::OnAccept(int nErrorCode){
	if (!nErrorCode){
		// [TPT]
		// Maella -Accurate measure of bandwidth: eDonkey data + control, network adapter-
		theApp.pBandWidthControl->AddeMuleInTCPOverall(0); // SYN
		// Maella end
		m_nPeningConnections++;
		if (m_nPeningConnections < 1){
			ASSERT ( false );
			m_nPeningConnections = 1;
		}
		if (TooManySockets(true) && !theApp.serverconnect->IsConnecting()){
			StopListening();
			return;
		}
		else if ( bListening == false )
			ReStartListening(); //If the client is still at maxconnections, this will allow it to go above it.. But if you don't, you will get a lowID on all servers.
	
		while (m_nPeningConnections){
			m_nPeningConnections--;
			CClientReqSocket* newclient = new CClientReqSocket(app_prefs);
			// [TPT]
			// Maella -Accurate measure of bandwidth: eDonkey data + control, network adapter-
			theApp.pBandWidthControl->AddeMuleOutTCPOverall(0);
			// Maella end
			if (!Accept(*newclient))
				newclient->Safe_Delete();
			else{
				newclient->AsyncSelect(FD_WRITE|FD_READ|FD_CLOSE);
				newclient->OnInit();
			}
			AddConnection();
		}
//		if (TooManySockets(true) && !theApp.serverconnect->IsConnecting())
//			StopListening();
	}
}

// [TPT]
// Maella -Code Improvement-
void CListenSocket::Process(){
	// Update counter
	m_OpenSocketsInterval = 0;
	opensockets = socket_list.GetSize();

	// Check state	
	for(POSITION pos = socket_list.GetHeadPosition(); pos != NULL; ){
		CClientReqSocket* cur_sock = socket_list.GetNext(pos);
		if(cur_sock->deletethis == true){
		   if(cur_sock->m_SocketData.hSocket != INVALID_SOCKET){
				cur_sock->Close();
			}
		   else {
				// Remove instance from the list (Recursive)
			   cur_sock->Delete_Timed();;
			}
		}
		else {
			cur_sock->CheckTimeOut();
		}
	}
	
	if((bListening == false) && 
	   (GetOpenSockets()+5 < app_prefs->GetMaxConnections() || theApp.serverconnect->IsConnecting() == true)){
	   ReStartListening();
	}
}
// Maella end

void CListenSocket::RecalculateStats(){
	MEMSET(m_ConnectionStates,0,sizeof m_ConnectionStates);
	for (POSITION pos = socket_list.GetHeadPosition(); pos != NULL; ){
		switch (socket_list.GetNext(pos)->GetConState()){
			case ES_DISCONNECTED:
				m_ConnectionStates[0]++;
				break;
			case ES_NOTCONNECTED:
				m_ConnectionStates[1]++;
				break;
			case ES_CONNECTED:
				m_ConnectionStates[2]++;
				break;
		}
   }
}

void CListenSocket::AddSocket(CClientReqSocket* toadd){
	socket_list.AddTail(toadd);
}

// Pichuei, se pueden abrir 2 sockets identicos? si no es asi hacer un socket_list.find
void CListenSocket::RemoveSocket(CClientReqSocket* todel){
	for (POSITION pos = socket_list.GetHeadPosition(); pos != NULL; ){
		POSITION posLast = pos;
		if ( socket_list.GetNext(pos) == todel )
			socket_list.RemoveAt(posLast);
   }
}

// [TPT] - Maella -Code Improvement-
void CListenSocket::KillAllSockets(){
	while(socket_list.GetHeadPosition() != NULL){
		CClientReqSocket* cur_socket = socket_list.GetHead();
		if (cur_socket->client)
			delete cur_socket->client;
		else
			delete cur_socket;
	}
}
// Maella end

void CListenSocket::AddConnection(){
	m_OpenSocketsInterval++;
	opensockets++;
}

bool CListenSocket::TooManySockets(bool bIgnoreInterval){
	if (GetOpenSockets() > app_prefs->GetMaxConnections() 
		|| (m_OpenSocketsInterval > (theApp.glob_prefs->GetMaxConperFive()*GetMaxConperFiveModifier()) && !bIgnoreInterval) ){
		return true;
	}
	else
		return false;
}

bool CListenSocket::IsValidSocket(CClientReqSocket* totest){
	return socket_list.Find(totest);
}

#ifdef _DEBUG
void CListenSocket::Debug_ClientDeleted(CUpDownClient* deleted){
	for (POSITION pos = socket_list.GetHeadPosition(); pos != NULL;){
		CClientReqSocket* cur_sock = socket_list.GetNext(pos);
		if (!AfxIsValidAddress(cur_sock, sizeof(CClientReqSocket))) {
			AfxDebugBreak(); 
		}
		if (_iDbgHeap >= 2)
			ASSERT_VALID(cur_sock);
		if (cur_sock->client == deleted){
			AfxDebugBreak();
		}
	}
}
#endif

void CListenSocket::UpdateConnectionsStatus()
{
	activeconnections = GetOpenSockets();
	if( peakconnections < activeconnections )
		peakconnections = activeconnections;
	// -khaos--+++>
	if (peakconnections>theApp.glob_prefs->GetConnPeakConnections())
		theApp.glob_prefs->Add2ConnPeakConnections(peakconnections);
	// <-----khaos-
	if( theApp.IsConnected() ){
		totalconnectionchecks++;
		float percent;
		percent = (float)(totalconnectionchecks-1)/(float)totalconnectionchecks;
		if( percent > .99f )
			percent = .99f;
		averageconnections = (averageconnections*percent) + (float)activeconnections*(1.0f-percent);
	}
}

float CListenSocket::GetMaxConperFiveModifier(){
	//This is a alpha test.. Will clean up for b version.
	const float SpikeSize = (float)GetOpenSockets() - averageconnections ;
	if ( SpikeSize < 1.0f )
		return 1.0f;

	// Remark: default factor 2.5, try to increase the number of connections per 5 second
	const float SpikeTolerance = 2.5f * (float)theApp.glob_prefs->GetMaxConperFive();
	if ( SpikeSize > SpikeTolerance )
		return 0.0f;
	
	return 1.0f - (SpikeSize/SpikeTolerance);
}